;==============================================
;  xmandel.asm
;
;    Requires MASM 6.0, 386/387 or i486, MS-DOS
;
;    Assembly language implementation of a fast
;    Mandelbrot generator, with extra features.
;
;    Copyright 1993 Scott Robert Ladd.
;    All Rights Reserved.
;==============================================

;-----------------------------------
; memory model and processor options
;-----------------------------------

.MODEL SMALL
.386
.387

;-----------------
; type definitions
;-----------------

PalEntry STRUCT
    Red   BYTE ?
    Green BYTE ?
    Blue  BYTE ?
PalEntry ENDS

;--------------
; stack segment
;--------------

.STACK 128

;--------------
; data segments
;--------------

.CONST

    XFactor   REAL8   0.003125 ; 1/320
    YFactor   REAL8   0.005000 ; 1/200

    ; for full mandelbrot
    ; XMin      REAL8  -2.3
    ; XMax      REAL8   0.9
    ; YMin      REAL8  -1.2
    ; YMax      REAL8   1.2

    XMin      REAL8  -0.64
    XMax      REAL8  -0.44
    YMin      REAL8   0.55
    YMax      REAL8   0.70

    Escape    REAL8   4.0

    XPixels   WORD    320
    YPixels   WORD    200
    MaxIter   WORD    150

    MyPalette PalEntry {63, 3, 3},{59, 5, 7},
                       {55, 6,11},{51, 8,15},
                       {47, 9,19},{43,11,23},
                       {39,13,27},{35,14,31},
                       {31,16,35},{27,17,39},
                       {23,19,43},{19,21,47},
                       {15,22,51},{11,24,55},
                       { 7,25,59},{ 3,27,63}

.DATA?
    DeltaX    REAL8   ?
    DeltaY    REAL8   ?
    RowVal    REAL8   ?
    Xv        REAL8   ?
    XTable    REAL8   200 DUP (?)
    OrigPal   BYTE     48 DUP (?)

;------------------------------
; code segment and main program
;------------------------------

.CODE

    .STARTUP

    ; reset the coprocessor
    fninit

    ; set 256-color, 320 by 200 display mode
    mov     ax,0013h
    int     10h

    ; save old palette
    push    ds
    pop     es
    mov     dx,OFFSET OrigPal
    mov     bx,1
    mov     cx,16
    mov     ax,1017h
    int     10h

    ; load my palette
    push    ds
    pop     es
    mov     dx,OFFSET MyPalette
    mov     bx,1
    mov     cx,16
    mov     ax,1012h
    int     10h

    ; set-up graphics segment (es) and index (bx)
    mov     ax,0A000h
    mov     es,ax

    ; calculate delta values
    ; DeltaX = (XMax - XMin) * XFactor
    fld     XMax
    fsub    XMin
    fmul    XFactor
    fstp    DeltaX

    ; DeltaY = (YMax - YMin) * YFactor
    fld     YMax
    fsub    YMin
    fmul    YFactor
    fstp    DeltaY

    ; fill array of Y values
    mov     di,2552

    ; load first element
    fld     XMax

SetXTable:
    fsub    DeltaX
    fst     XTable[di]

    and     di,di
    jz      Setup

    sub     di,8
    jmp     SetXTable

Setup:
    fstp    st     ; pop top of stack

    ; set RowVal = YMin
    fld     YMax
    fstp    RowVal

    ; set row (dx) to 0
    mov     dx,0

    ; set bx to point to first pixel
    mov     bx,0

RowLoop:
    ; set col (si) to zero
    mov     si,0

ColLoop:
    ; Xv = XTable[dx * 8]
    fld     RowVal      ; initial Y
    mov     di,si
    shl     di,3
    fld     XTable[di]
    fst     Xv          ; stays on stack as X

    mov     cx,0        ; cx contains color value

InnerLoop:
    ; put new squares on stack
    fld     st(1)
    fmul    st,st
    fld     st(1)
    fmul    st,st

    ; if (Xsq + Ysq) > Escape, plot pixel
    fld     st(1)
    fadd    st,st(1)
    fcomp   Escape
    fnstsw  ax
    test    ax,8100h
    jz      ClearAndPlot

    ; Y = 2 * (Y * X) + RowVal
    fld     st(3)
    fmul    st,st(3)
    fadd    st,st
    fadd    RowVal
    fstp    st(4)

    ; X = Xsq - Ysq + Xv
    fld     st
    fsub    st,st(2)
    fadd    Xv
    fstp    st(3)

    ; remove old squares from stack
    fstp    st
    fstp    st

    ; increment count
    inc     cx

    ; if color >= MaxIter, plot the pixel
    cmp     cx,MaxIter
    jl      InnerLoop

    mov     BYTE PTR es:[bx],0
    jmp     PlotPixel

ClearAndPlot:
    ; remove squares from stack
    fstp    st
    fstp    st

    ; set color
    and     cx,15
    inc     cx
    mov     es:[bx],cl

PlotPixel:
    ; remove X and Y from stack
    fstp    st
    fstp    st

    ; move to next pixel row
    inc     bx

    ; increment column and loop
    inc     si
    cmp     si,XPixels
    jl      ColLoop

    ; increment ColVal
    fld     RowVal
    fsub    DeltaY
    fstp    RowVal

    ; increment row and loop
    inc     dx
    cmp     dx,YPixels
    jl      RowLoop

KeyLoop:
    ; wait for keypress
    mov     ah,1
    int     16h
    jz      KeyLoop

    ; eat keystroke
    mov     ah,0
    int     16h
    
    ; restore old palette
    push    ds
    pop     es
    mov     dx,OFFSET OrigPal
    mov     bx,1
    mov     cx,16
    mov     ax,1012h
    int     10h

    ; set text display mode
    mov     ax,0003h
    int     10h

    mov     ax,4C00h
    int     21h

END

